// Copyright 2025 "Johan Maasing" <johan@zoom.nu>
//
//   Licensed under the Apache License, Version 2.0 (the "License");
//   you may not use this file except in compliance with the License.
//   You may obtain a copy of the License at
//
//       http://www.apache.org/licenses/LICENSE-2.0
//
//   Unless required by applicable law or agreed to in writing, software
//   distributed under the License is distributed on an "AS IS" BASIS,
//   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//   See the License for the specific language governing permissions and
//   limitations under the License.
package nu.zoom.dsl.freemarker;

import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;
import nu.zoom.dsl.ast.DocumentNode;
import nu.zoom.dsl.ast.GeneratorNode;
import nu.zoom.dsl.ast.Meta;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class Generator {
	private final Path templatesDir;
	private final DocumentNode documentNode;
	private final Path outputDir;
	private final Configuration cfg;
	private final String TEMPLATE_EXTENSION = ".ftl";
	private final int TEMPLATE_EXTENSION_LENGTH = TEMPLATE_EXTENSION.length();

	public Generator(Path templatesDir, DocumentNode documentNode, Path outputDir) throws IOException {
		this.templatesDir = Objects.requireNonNull(templatesDir);
		this.documentNode = Objects.requireNonNull(documentNode);
		this.outputDir = Objects.requireNonNull(outputDir);
		this.cfg = new Configuration(Configuration.VERSION_2_3_34);
		cfg.setDirectoryForTemplateLoading(templatesDir.toFile());
		cfg.setDefaultEncoding("UTF-8");
		cfg.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);
		cfg.setLogTemplateExceptions(false);
		cfg.setWrapUncheckedExceptions(true);
		cfg.setFallbackOnNullLoopVariable(false);
	}

	public List<Path> generate() throws IOException, TemplateException {
		try (Stream<Path> files = Files.walk(templatesDir)) {
			List<Path> templates = files
				.filter(p -> {
						var fname = p.getFileName().toString();
						return fname.length() > TEMPLATE_EXTENSION_LENGTH && fname.endsWith(TEMPLATE_EXTENSION);
					}
				)
				.map(p -> templatesDir.relativize(p))
				.toList();
			ArrayList<Path> out = new ArrayList<>();
			for (Path template : templates) {
				Path outpath = outputDir.resolve(outputFilenameFromTemplate(template.toString()));
				Files.createDirectories(outpath.getParent());
				Path templateSubdirectory = template.getParent();
				ArrayList<String> templateDirectories= new ArrayList<>() ;
				if (templateSubdirectory != null) {
					var ti = templateSubdirectory.iterator();
					while (ti.hasNext()) {
						templateDirectories.add(ti.next().toString());
					}
				}
				String templateName = template.getFileName().toString();
				Meta meta = new Meta(templateDirectories, templateName);
				GeneratorNode generatorNode = new GeneratorNode(
					this.documentNode.config(),
					this.documentNode.typeDefinitions(),
					this.documentNode.endpoints(),
					this.documentNode.states(),
					meta
				);
				Template ftl = this.cfg.getTemplate(template.toString());
				try (var outw = Files.newBufferedWriter(outpath, StandardCharsets.UTF_8)) {
					ftl.process(generatorNode, outw);
					out.add(outpath);
				}
			}
			return out;
		}
	}

	private String outputFilenameFromTemplate(String template) {
		return template.substring(0, template.length() - TEMPLATE_EXTENSION_LENGTH);
	}
}
